# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The main game object.

define (require, exports, module) ->
    GraphicsDevice = require './graphics/GraphicsDevice'

    $ = require 'jquery'
    _ = require 'underscore'

    # `requestAnimationFrame` polyfill.
    _requestAnimationFrame = (callback) ->
        @setTimeout callback, 1000.0 / 60.0

    # The Game object, which manages everything and more.
    Game = (window, opts) ->
        @opts = opts

        # If the game is running.
        @_running = false

        # The total elapsed time.
        @_lastTime = (new Date).getTime()
        @_elapsed = 0.0

        # The DOM window object.
        @window = window

        # The DOM canvas object.
        @canvas = window.document.createElement 'canvas'

        # The graphics device binding.
        @g = new GraphicsDevice @canvas, opts.graphics

        # Set up possible redirections for _requestAnimationFrame.
        @_requestAnimationFrame = _(((f) ->
            () -> f.apply @, arguments
        )(
            @window.requestAnimationFrame or
            @window.webkitRequestAnimationFrame or
            @window.mozRequestAnimationFrame or
            @window.oRequestAnimationFrame or
            @window.msRequestAnimationFrame or
            _requestAnimationFrame
        )).bind @window

        @_setupDOM()

        return

    # Bind various DOM events.
    Game::_setupDOM = ->
        # If we've set fit to parent, then fit to parent on resize.
        if @opts.fitToParent
            $(window).bind 'resize', _(_(@_refit).bind @).debounce 300
        return

    Game::_refit = () ->
        $parent = $(@canvas).parent()
        @canvas.width = $parent.width()
        @canvas.height = $parent.height()

    # Trampoline game steps.
    Game::_trampoline = ->
        if @_running
            @_requestAnimationFrame _(@_trampoline).bind @

        # Do timing calculations.
        endTime = (new Date).getTime()
        dt = endTime - @_lastTime
        @_elapsed += dt
        @_lastTime = endTime

        # Pass them off to the tick function.
        @tick dt, @_elapsed

    # Run the game.
    Game::run = ->
        if @opts.fitToParent
            @_refit()

        @_running = true

        @initialize()
        @_trampoline()

    # Perform any initialization logic. This is up to the implementer.
    Game::initialize = ->
        return

    # Run the draw/update loop of the game. This is up to the implementer.
    Game::tick = (dt, elapsed) ->
        return

    module.exports = Game
