# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is a deferred-driven content loading system.

define (require, exports, module) ->
    $ = require 'jquery'
    _ = require 'underscore'

    Material = require './graphics/Material'
    Mesh = require './graphics/Mesh'
    Color = require './graphics/Color'

    # Load an image using a deferred.
    loadImage = (src) ->
        deferred = $.Deferred()
        img = new Image
        img.onload = () -> deferred.resolve img
        img.onerror = deferred.reject
        img.src = src
        deferred.promise()

    # Load a text file via AJAX using a deferred.
    loadXhr = (src, type) ->
        $.get(src, null, null, type).pipe \
            (data) -> data
            ,
            (e) ->
                # TODO: Throwing a new kind of error might be bad; find another
                #       exception class (or maybe add a new one?).
                name    : 'HttpError'
                message : e.statusText
                error   : e

    # Load a shader using a deferred.
    loadShader = (g, type, src) ->
        loadXhr(src, 'text').pipe (data) ->
            try
                g.compileShader type, data
            catch e
                $.Deferred().reject e

    # Load a 2D texture using a deferred.
    loadTexture2D = (g, src, levels, internalFormat, format, type) ->
        loadImage(src).pipe (img) ->
            try
                g.createTexture2D img, levels, internalFormat, format, type
            catch e
                $.Deferred().reject e

    # Load a mesh (using the web-worker-compatible format).
    loadMesh = (g, src) ->
        deferred = $.Deferred()
        worker = new window.Worker src
        worker.onmessage = (event) ->
            # We should compile these to buffers.
            data = event.data

            vertices = g.createBuffer g.gl.ARRAY_BUFFER, data.vertices, g.gl.STATIC_DRAW
            indices = if data.indices? then g.createBuffer(g.gl.ARRAY_BUFFER, data.indices, g.gl.STATIC_DRAW) else null
            stcoords = if data.stcoords? then g.createBuffer(g.gl.ARRAY_BUFFER, data.stcoords, g.gl.STATIC_DRAW) else null
            normals = if data.normals? then g.createBuffer(g.gl.ARRAY_BUFFER, data.normals, g.gl.STATIC_DRAW) else null
            tangents = if data.tangents? then g.createBuffer(g.gl.ARRAY_BUFFER, data.tangents, g.gl.STATIC_DRAW) else null

            mesh = new Mesh vertices, indices, stcoords, normals, tangents
            deferred.resolve mesh
        worker.onerror = (e) ->
            deferred.reject e
        deferred.promise()

    # Load a material (via XHR JSON).
    loadMaterial = (g, src) ->
        loadXhr(src, "json").pipe (data) ->
            try
                ambient = Color.fromArray data.ambient
                diffuse = Color.fromArray data.diffuse
                specular = Color.fromArray data.specular
                emission = Color.fromArray data.emission

                if data.texture?
                    loadTexture2D(g, data.texture, 0, g.gl.RGBA, g.gl.RGBA, g.gl.UNSIGNED_BYTE).pipe (texture) ->
                        try
                            new Material ambient, diffuse, specular, emission,
                                         data.shininess, texture
                        catch e
                            $.Deferred().reject e
                else
                    new Material ambient, diffuse, specular, emission,
                                 data.shininess
            catch e
                $.Deferred().reject e

    module.exports =
        loadImage       : loadImage
        loadXhr         : loadXhr
        loadShader      : loadShader
        loadTexture2D   : loadTexture2D
        loadMesh        : loadMesh
        loadMaterial    : loadMaterial
