# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation of the hash-based message authentication code algorithm.

define (require, exports, module) ->
    bit = require './../math/bit'

    SHA1 = require './hash/SHA1'
    SHA256 = require './hash/SHA256'
    MD5 = require './hash/MD5'

    # The HMAC algorithm requires an underlying `algorithm` (such as
    # `basilisk.crypto.hash.SHA256`) that exposes an appropriate interface. It
    # also requires a `key` (which may only be set once). It can also be
    # initialized with an initial `input` value.
    HMAC = (algorithm, key, input='') ->
        # If the algorithm is null, cheaply create an HMAC object without
        # initializing anything.
        if algorithm is null
            return

        # Turn the key into a bytes array, if we have to.
        if typeof key == 'string'
            key = bit.stringToBytes key

        # As per RFC 2104, if the key is longer than the block size, then use
        # the hashing algorithm to reduce it to the block size.
        if key.length > algorithm.blockSize
            key = (new algorithm key).digest()

        # Construct the outer key padding by XORing the key with 0x5c5c...5c5c.
        outerKeyPad = new bit.ByteArray algorithm.blockSize
        for i in [0..algorithm.blockSize - 1]
            outerKeyPad[i] = 0x5c ^ key[i]

        # Construct the inner key padding by XORing the key with 0x3636...3636.
        innerKeyPad = new bit.ByteArray algorithm.blockSize
        for i in [0..algorithm.blockSize - 1]
            innerKeyPad[i] = 0x36 ^ key[i]

        # Create an inner digester.
        @inner = new algorithm innerKeyPad

        # Create an outer digester.
        @outer = new algorithm outerKeyPad

        @update input
        return

    # Get the hash object for the current state.
    HMAC::_current = ->
        @outer.copy().update @inner.digest()

    # Digest all input and output the hash.
    HMAC::digest = ->
        @_current().digest()

    # Digest all input and output the hex representation of the hash.
    HMAC::hexdigest = ->
        @_current().hexdigest()

    # Update the HMAC algorithm with more input.
    HMAC::update = (input) ->
        @inner.update input
        @

    # Copy the HMAC object.
    HMAC::copy = ->
        hmac = new HMAC null

        hmac.inner = @inner.copy()
        hmac.outer = @outer.copy()
        hmac

    # Create a HMAC object, using the SHA1 algorithm.
    HMAC.sha1new = (key, input='') ->
        new HMAC SHA1, key, input

    # Create a HMAC object, using the SHA256 algorithm.
    HMAC.sha256new = (key, input='') ->
        new HMAC SHA256, key, input

    # Create a HMAC object, using the MD5256 algorithm.
    HMAC.md5new = (key, input='') ->
        new HMAC MD5, key, input

    module.exports = HMAC
