# Copyright 2011 Perry Tan
#                Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation of the MD5 hashing algorithm.

define (require, exports, module) ->
    lang = require './../../lang'
    bit = require './../../math/bit'

    # Use the Merkle-Damgard construction.
    _MerkleDamgard = require './_MerkleDamgard'

    # Alias rol32 to bit.rol32.
    rol32 = bit.rol32

    # Create a MD5 hash object, which calculates the hash of a given input.
    MD5 = (input='') ->
        # State variables.
        @_h = new bit.Word32Array [
            0x67452301,
            0xefcdab89,
            0x98badcfe,
            0x10325476
        ]

        _MerkleDamgard.call @, input
        return

    # The table of constants constructed from the sine function.
    MD5.T = new bit.Word32Array [
        0xd76aa478, 0xe8c7b756, 0x242070db, 0xc1bdceee, 0xf57c0faf, 0x4787c62a,
        0xa8304613, 0xfd469501, 0x698098d8, 0x8b44f7af, 0xffff5bb1, 0x895cd7be,
        0x6b901122, 0xfd987193, 0xa679438e, 0x49b40821, 0xf61e2562, 0xc040b340,
        0x265e5a51, 0xe9b6c7aa, 0xd62f105d, 0x02441453, 0xd8a1e681, 0xe7d3fbc8,
        0x21e1cde6, 0xc33707d6, 0xf4d50d87, 0x455a14ed, 0xa9e3e905, 0xfcefa3f8,
        0x676f02d9, 0x8d2a4c8a, 0xfffa3942, 0x8771f681, 0x6d9d6122, 0xfde5380c,
        0xa4beea44, 0x4bdecfa9, 0xf6bb4b60, 0xbebfbc70, 0x289b7ec6, 0xeaa127fa,
        0xd4ef3085, 0x04881d05, 0xd9d4d039, 0xe6db99e5, 0x1fa27cf8, 0xc4ac5665,
        0xf4292244, 0x432aff97, 0xab9423a7, 0xfc93a039, 0x655b59c3, 0x8f0ccc92,
        0xffeff47d, 0x85845dd1, 0x6fa87e4f, 0xfe2ce6e0, 0xa3014314, 0x4e0811a1,
        0xf7537e82, 0xbd3af235, 0x2ad7d2bb, 0xeb86d391
    ]

    # The list of per-round shift amounts.
    MD5.R = new bit.Word32Array [
        7, 12, 17, 22, 7, 12, 17, 22, 7, 12, 17, 22, 7, 12, 17, 22,
        5,  9, 14, 20, 5,  9, 14, 20, 5,  9, 14, 20, 5,  9, 14, 20,
        4, 11, 16, 23, 4, 11, 16, 23, 4, 11, 16, 23, 4, 11, 16, 23,
        6, 10, 15, 21, 6, 10, 15, 21, 6, 10, 15, 21, 6, 10, 15, 21
    ]

    # Inherit from _MerkleDamgard.
    MD5:: = lang.object _MerkleDamgard::
    MD5::constructor = MD5

    # Perform a round of MD5 hashing.
    MD5::_performRound = ->
        # Break the chunk into 16 32-bit little endian words.
        w = bit.bytesToWords32LE @_chunk, new bit.Word32Array(16)

        # Hash values for the chunk
        [a, b, c, d] = @_h

        # Main loop
        for i in [0..63]
            if 0 <= i <= 15 
                f = (b & c) | (~b & d)
                g = i
            else if 16 <= i <= 31
                f = (d & b) | (~d & c)
                g = (5 * i + 1) % 16
            else if 32 <= i <= 47
                f = b ^ c ^ d
                g = (3 * i + 5) % 16
            else if 48 <= i <= 63
                f = c ^ (b | ~d)
                g = (7 * i) % 16

            temp = d
            d = c
            c = b
            b = (b + rol32 (a + f + MD5.T[i] + w[g]), MD5.R[i]) & 0xffffffff
            a = temp

        @_h[0] = (@_h[0] + a) & 0xffffffff
        @_h[1] = (@_h[1] + b) & 0xffffffff
        @_h[2] = (@_h[2] + c) & 0xffffffff
        @_h[3] = (@_h[3] + d) & 0xffffffff

        return 

    # Calculate a MD5 hash using the input the object was given.
    MD5::_digest = ->
        # Store the current message length, in bits.
        n = @_n * 8

        # Pad 0x80.
        @update new bit.ByteArray [ 0x80 ]

        if @_chunkLength > 56
            # This block is dead, force an update already.
            @_update()

        # Add the length to the end of the current chunk as a 64-bit
        # little-endian number.
        i = 56
        while n != 0
            @_chunk[i] = n & 0xff
            n >>= 8
            i++

        @_update()

        bit.words32LEToBytes @_h

    # Expose block size (in bytes) for things like HMAC to use.
    MD5.blockSize = 64

    module.exports = MD5
