# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation of the SHA1 hashing algorithm.

define (require, exports, module) ->
    lang = require './../../lang'
    bit = require './../../math/bit'

    # Use the Merkle-Damgard construction.
    _MerkleDamgard = require './_MerkleDamgard'

    # Alias rol32 to bit.rol32.
    rol32 = bit.rol32

    # Create a SHA1 hash object, which calculates the hash of a given input.
    SHA1 = (input='') ->
        # The hash variables.
        @_h = new bit.Word32Array [
            0x67452301,
            0xefcdab89,
            0x98badcfe,
            0x10325476,
            0xc3d2e1f0
        ]

        _MerkleDamgard.call @, input
        return

    # Inherit from _MerkleDamgard (hopefully without exploding).
    SHA1:: = lang.object _MerkleDamgard::
    SHA1::constructor = SHA1

    # Perform a round of SHA1 hashing.
    SHA1::_performRound = ->
        # Break the chunk into 16 32-bit big endian words.
        w = bit.bytesToWords32BE @_chunk, new bit.Word32Array(80)

        # Extend the 16 32-bit words into 80 32-bit words.
        for i in [16..79]
            w[i] = rol32(w[i - 3] ^ w[i - 8] ^ w[i - 14] ^ w[i - 16], 1)

        # Initialize hash values.
        [a, b, c, d, e] = @_h

        # Main loop.
        for i in [0..79]
            if 0 <= i <= 19
                f = (b & c) | (~b & d)
                k = 0x5a827999
            else if 20 <= i <= 39
                f = b ^ c ^ d
                k = 0x6ed9eba1
            else if 40 <= i <= 59
                f = (b & c) | (b & d) | (c & d)
                k = 0x8f1bbcdc
            else if 60 <= i <= 79
                f = b ^ c ^ d
                k = 0xca62c1d6

            temp = (rol32(a, 5) + f + e + k + w[i]) & 0xffffffff
            e = d
            d = c
            c = rol32(b, 30)
            b = a
            a = temp

        @_h[0] = (@_h[0] + a) & 0xffffffff
        @_h[1] = (@_h[1] + b) & 0xffffffff
        @_h[2] = (@_h[2] + c) & 0xffffffff
        @_h[3] = (@_h[3] + d) & 0xffffffff
        @_h[4] = (@_h[4] + e) & 0xffffffff

        return

    # Calculate a SHA1 hash using the input the object was given.
    SHA1::_digest = ->
        # Store the current message length, in bits.
        n = @_n * 8

        # Pad 0x80.
        @update new bit.ByteArray [ 0x80 ]

        if @_chunkLength > 56
            # This block is dead, force an update already.
            @_update()

        # Add the length to the end of the current chunk as a 64-bit big-endian
        # number.
        i = 63
        while n != 0
            @_chunk[i] = n & 0xff
            n >>= 8
            i--

        @_update()

        bit.words32BEToBytes @_h

    # Expose block size (in bytes) for things like HMAC to use.
    SHA1.blockSize = 64

    module.exports = SHA1
