# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Implementation of the SHA256 hashing algorithm.

define (require, exports, module) ->
    lang = require './../../lang'
    bit = require './../../math/bit'

    # Use the Merkle-Damgard construction.
    _MerkleDamgard = require './_MerkleDamgard'

    # Alias ror32 to bit.ror32.
    ror32 = bit.ror32

    # Create a SHA256 hash object, which calculates the hash of a given input.
    SHA256 = (input='') ->
        # The first 32-bits of the fractional parts of the square roots of the
        # first 8 primes from 2 to 19.
        @_h = new bit.Word32Array [
            0x6a09e667,
            0xbb67ae85,
            0x3c6ef372,
            0xa54ff53a,
            0x510e527f,
            0x9b05688c,
            0x1f83d9ab,
            0x5be0cd19
        ]

        _MerkleDamgard.call @, input
        return

    # The table of round constants, i.e. the first 32-bits of the cube roots of
    # the first 64 primes from 2 to 311.
    SHA256.K = new bit.Word32Array [
        0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1,
        0x923f82a4, 0xab1c5ed5, 0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
        0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174, 0xe49b69c1, 0xefbe4786,
        0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
        0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147,
        0x06ca6351, 0x14292967, 0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
        0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85, 0xa2bfe8a1, 0xa81a664b,
        0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
        0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a,
        0x5b9cca4f, 0x682e6ff3, 0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
        0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
    ]

    # Inherit from _MerkleDamgard (hopefully without exploding).
    SHA256:: = lang.object _MerkleDamgard::
    SHA256::constructor = SHA256

    # Perform a round of SHA256 hashing.
    SHA256::_performRound = ->
        # Break the chunk into 16 32-bit big endian words.
        w = bit.bytesToWords32BE @_chunk, new bit.Word32Array(64)

        # Extend the 16 32-bit words into 64 32-bit words.
        for i in [16..63]
            s0 = ror32(w[i - 15], 7) ^ ror32(w[i - 15], 18) ^ (w[i - 15] >>> 3)
            s1 = ror32(w[i - 2], 17) ^ ror32(w[i - 2], 19) ^ (w[i - 2] >>> 10)

            w[i] = (s1 + w[i - 7] + s0 + w[i - 16]) & 0xffffffff

        # Initialize hash values.
        [a, b, c, d, e, f, g, h] = @_h

        # Main loop.
        for i in [0..63]
            ch = (e & f) ^ (~e & g)
            maj = (a & b) ^ (a & c) ^ (b & c)

            s0 = ror32(a, 2) ^ ror32(a, 13) ^ ror32(a, 22)
            s1 = ror32(e, 6) ^ ror32(e, 11) ^ ror32(e, 25)

            t1 = (h + s1 + ch + SHA256.K[i] + w[i]) & 0xffffffff
            t2 = (s0 + maj) & 0xffffffff

            h = g
            g = f
            f = e
            e = (d + t1) & 0xffffffff
            d = c
            c = b
            b = a
            a = (t1 + t2) & 0xffffffff

        @_h[0] = (@_h[0] + a) & 0xffffffff
        @_h[1] = (@_h[1] + b) & 0xffffffff
        @_h[2] = (@_h[2] + c) & 0xffffffff
        @_h[3] = (@_h[3] + d) & 0xffffffff
        @_h[4] = (@_h[4] + e) & 0xffffffff
        @_h[5] = (@_h[5] + f) & 0xffffffff
        @_h[6] = (@_h[6] + g) & 0xffffffff
        @_h[7] = (@_h[7] + h) & 0xffffffff

        return

    # Calculate a SHA256 hash using the input the object was given.
    SHA256::_digest = ->
        # Store the current message length, in bits.
        n = @_n * 8

        # Pad 0x80.
        @update new bit.ByteArray [ 0x80 ]

        if @_chunkLength > 56
            # This block is dead, force an update already.
            @_update()

        # Add the length to the end of the current chunk as a 64-bit big-endian
        # number.
        i = 63
        while n != 0
            @_chunk[i] = n & 0xff
            n >>= 8
            i--

        @_update()

        bit.words32BEToBytes @_h

    # Expose block size (in bytes) for things like HMAC to use.
    SHA256.blockSize = 64

    module.exports = SHA256
