# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Various bitwise operations.

define (require, exports, module) ->
    types = require './../types'

    # Perform a 32-bit right rotation of `x` by `n`.
    ror32 = (x, n) ->
        (x >>> n) | (x << (32 - n)) & 0xffffffff

    # Perform a 32-bit left rotation of `x` by `n`.
    rol32 = (x, n) ->
        (x << n) | (x >>> (32 - n)) & 0xffffffff

    # This is the type for storing 32-bit words.
    Word32Array = types.Int32Array

    # This is the type for storing raw bytes.
    ByteArray = types.Int8Array

    # Convert a JavaScript string into an array of raw characters (i.e. no
    # Unicode characters).
    stringToBytes = (str) ->
        bytes = []
        if str.length != 0
            for i in [0..str.length - 1]
                ord = str.charCodeAt i
                seq = []
                while true
                    seq.push ord & 0xff
                    ord >>= 8
                    if ord == 0
                        break
                bytes = bytes.concat seq.reverse()
        new ByteArray bytes

    # Convert an array of `bytes` to an array of 32-bit big-endian words.
    bytesToWords32BE = (bytes, words) ->
        words = new Word32Array(Math.ceil(bytes.length / 4)) if not words?

        for i in [0..bytes.length - 1]
            b = i * 8
            words[b >>> 5] |= (bytes[i] & 0xff) << (24 - b % 32)

        words

    # Convert an array of `bytes` to an array of 32-bit little-endian words.
    bytesToWords32LE = (bytes, words) ->
        words = new Word32Array(Math.ceil(bytes.length / 4)) if not words?

        for i in [0..bytes.length - 1]
            b = i * 8
            words[b >>> 5] |= (bytes[i] & 0xff) << (b % 32)

        words

    # Convert an array of 32-bit big-endian `words` to bytes.
    words32BEToBytes = (words, bytes) ->
        bytes = new ByteArray(4 * words.length) if not bytes?

        for i in [0..words.length - 1]
            word = words[i]
            b = i * 4

            bytes[b + 3] = word & 0xff
            word >>= 8

            bytes[b + 2] = word & 0xff
            word >>= 8

            bytes[b + 1] = word & 0xff
            word >>= 8

            bytes[b    ] = word & 0xff

        bytes

    # Convert an array of 32-bit little-endian `words` to bytes.
    words32LEToBytes = (words, bytes) ->
        bytes = new ByteArray(4 * words.length) if not bytes?

        for i in [0..words.length - 1]
            word = words[i]
            b = i * 4

            bytes[b   ] = word & 0xff
            word >>= 8

            bytes[b + 1] = word & 0xff
            word >>= 8

            bytes[b + 2] = word & 0xff
            word >>= 8

            bytes[b + 3] = word & 0xff

        bytes

    module.exports =
        ror32           : ror32
        rol32           : rol32
        Word32Array     : Word32Array
        ByteArray       : ByteArray
        stringToBytes   : stringToBytes
        bytesToWords32BE: bytesToWords32BE
        bytesToWords32LE: bytesToWords32LE
        words32BEToBytes: words32BEToBytes
        words32LEToBytes: words32LEToBytes
