# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is the Matrix4 class, which represents a four-by-four square matrix.

define (require, exports, module) ->
    types = require './../../types'
    exceptions = require './../../exceptions'

    Vector3 = require './Vector3'
    Quaternion = require './Quaternion'

    # A representation of a fourth order square matrix.
    Matrix4 = (m) ->
        @m = m
        return

    # Alias `types.Float32Array`.
    Matrix4.Array = types.Float32Array

    # Create a matrix with arguments in row-major.
    Matrix4.fromRowMajor = (m00=0, m01=0, m02=0, m03=0, m10=0, m11=0, m12=0, m13=0, m20=0, m21=0, m22=0, m23=0, m30=0, m31=0, m32=0, m33=0) ->
        new Matrix4 new Matrix4.Array [
            m00, m10, m20, m30,
            m01, m11, m21, m31,
            m02, m12, m22, m32,
            m03, m13, m23, m33
        ]

    # Create a matrix with arguments in column-major.
    Matrix4.fromColumnMajor = (m00=0, m10=0, m20=0, m30=0, m01=0, m11=0, m21=0, m31=0, m02=0, m12=0, m22=0, m32=0, m03=0, m13=0, m23=0, m33=0) ->
        new Matrix4 new Matrix4.Array [
            m00, m10, m20, m30,
            m01, m11, m21, m31,
            m02, m12, m22, m32,
            m03, m13, m23, m33
        ]

    # Represent the `Matrix4` in string format.
    Matrix4::toString = ->
        "Matrix4([" + @m[ 0] + ", " + @m[ 1] + ", " + @m[ 2] + ", " + @m[ 3] + "," + "\n" +
        "         " + @m[ 4] + ", " + @m[ 5] + ", " + @m[ 6] + ", " + @m[ 7] + "," + "\n" +
        "         " + @m[ 8] + ", " + @m[ 9] + ", " + @m[10] + ", " + @m[11] + "," + "\n" +
        "         " + @m[12] + ", " + @m[13] + ", " + @m[14] + ", " + @m[15] + "])"

    # Copy the `Matrix4` into a new `Matrix4`. As with `Vector3::copy`, this
    # allows the copying of a `Matrix4` the original can be left untouched.
    Matrix4::copy = ->
        new Matrix4 new Matrix4.Array @m

    # Check if this `Matrix4` is exactly equal to another `Matrix4` `rhs`.
    Matrix4::equals = (rhs) ->
        !!(
            @m[ 0] == rhs.m[ 0] &&
            @m[ 4] == rhs.m[ 4] &&
            @m[ 8] == rhs.m[ 8] &&
            @m[12] == rhs.m[12] &&
            @m[ 1] == rhs.m[ 1] &&
            @m[ 5] == rhs.m[ 5] &&
            @m[ 9] == rhs.m[ 9] &&
            @m[13] == rhs.m[13] &&
            @m[ 2] == rhs.m[ 2] &&
            @m[ 6] == rhs.m[ 6] &&
            @m[10] == rhs.m[10] &&
            @m[14] == rhs.m[14] &&
            @m[ 3] == rhs.m[ 3] &&
            @m[ 7] == rhs.m[ 7] &&
            @m[11] == rhs.m[11] &&
            @m[15] == rhs.m[15]
        )

    # Add this `Matrix4` to another `Matrix4` `rhs` in-place.
    Matrix4::iadd = (rhs) ->
        @m[ 0] += rhs.m[ 0]
        @m[ 4] += rhs.m[ 4]
        @m[ 8] += rhs.m[ 8]
        @m[12] += rhs.m[12]

        @m[ 1] += rhs.m[ 1]
        @m[ 5] += rhs.m[ 5]
        @m[ 9] += rhs.m[ 9]
        @m[13] += rhs.m[13]

        @m[ 2] += rhs.m[ 2]
        @m[ 6] += rhs.m[ 6]
        @m[10] += rhs.m[10]
        @m[14] += rhs.m[14]

        @m[ 3] += rhs.m[ 3]
        @m[ 7] += rhs.m[ 7]
        @m[11] += rhs.m[11]
        @m[15] += rhs.m[15]

        @

    # Add this `Matrix4` to another `Matrix4` `rhs`, creating a new Matrix4.
    Matrix4::add = (rhs) ->
        @copy().iadd rhs

    # Multiply this `Matrix4` with something in-place.
    Matrix4::imul = (rhs) ->
        if typeof rhs == 'number'
            @ismul rhs
        else
            @immul rhs

    # Multiply this `Matrix4` with something.
    Matrix4::mul = (rhs) ->
        @copy().imul rhs

    # Multpliy this `Matrix4` by another `Matrix4` `rhs` in-place.
    Matrix4::immul = (rhs) ->
        [
            @m[ 0], @m[ 4], @m[ 8], @m[12], 
            @m[ 1], @m[ 5], @m[ 9], @m[13],
            @m[ 2], @m[ 6], @m[10], @m[14],
            @m[ 3], @m[ 7], @m[11], @m[15]
        ] = [
            @m[ 0] * rhs.m[ 0] + @m[ 4] * rhs.m[ 1] + @m[ 8] * rhs.m[ 2] + @m[12] * rhs.m[ 3],
            @m[ 0] * rhs.m[ 4] + @m[ 4] * rhs.m[ 5] + @m[ 8] * rhs.m[ 6] + @m[12] * rhs.m[ 7],
            @m[ 0] * rhs.m[ 8] + @m[ 4] * rhs.m[ 9] + @m[ 8] * rhs.m[10] + @m[12] * rhs.m[11],
            @m[ 0] * rhs.m[12] + @m[ 4] * rhs.m[13] + @m[ 8] * rhs.m[14] + @m[12] * rhs.m[15],

            @m[ 1] * rhs.m[ 0] + @m[ 5] * rhs.m[ 1] + @m[ 9] * rhs.m[ 2] + @m[13] * rhs.m[ 3],
            @m[ 1] * rhs.m[ 4] + @m[ 5] * rhs.m[ 5] + @m[ 9] * rhs.m[ 6] + @m[13] * rhs.m[ 7],
            @m[ 1] * rhs.m[ 8] + @m[ 5] * rhs.m[ 9] + @m[ 9] * rhs.m[10] + @m[13] * rhs.m[11],
            @m[ 1] * rhs.m[12] + @m[ 5] * rhs.m[13] + @m[ 9] * rhs.m[14] + @m[13] * rhs.m[15],

            @m[ 2] * rhs.m[ 0] + @m[ 6] * rhs.m[ 1] + @m[10] * rhs.m[ 2] + @m[14] * rhs.m[ 3],
            @m[ 2] * rhs.m[ 4] + @m[ 6] * rhs.m[ 5] + @m[10] * rhs.m[ 6] + @m[14] * rhs.m[ 7],
            @m[ 2] * rhs.m[ 8] + @m[ 6] * rhs.m[ 9] + @m[10] * rhs.m[10] + @m[14] * rhs.m[11],
            @m[ 2] * rhs.m[12] + @m[ 6] * rhs.m[13] + @m[10] * rhs.m[14] + @m[14] * rhs.m[15],

            @m[ 3] * rhs.m[ 0] + @m[ 7] * rhs.m[ 1] + @m[11] * rhs.m[ 2] + @m[15] * rhs.m[ 3],
            @m[ 3] * rhs.m[ 4] + @m[ 7] * rhs.m[ 5] + @m[11] * rhs.m[ 6] + @m[15] * rhs.m[ 7],
            @m[ 3] * rhs.m[ 8] + @m[ 7] * rhs.m[ 9] + @m[11] * rhs.m[10] + @m[15] * rhs.m[11],
            @m[ 3] * rhs.m[12] + @m[ 7] * rhs.m[13] + @m[11] * rhs.m[14] + @m[15] * rhs.m[15]
        ]
        @

    # Multiply this `Matrix4` by another `Matrix4` `rhs`, creating a new
    # Matrix4.
    Matrix4::mmul = (rhs) ->
        @copy().immul rhs

    # Multpliy this `Matrix4` by a scalar value `rhs` in-place.
    Matrix4::ismul = (rhs) ->
        @m[ 0] *= rhs
        @m[ 4] *= rhs
        @m[ 8] *= rhs
        @m[12] *= rhs
        @m[ 1] *= rhs
        @m[ 5] *= rhs
        @m[ 9] *= rhs
        @m[13] *= rhs
        @m[ 2] *= rhs
        @m[ 6] *= rhs
        @m[10] *= rhs
        @m[14] *= rhs
        @m[ 3] *= rhs
        @m[ 7] *= rhs
        @m[11] *= rhs
        @m[15] *= rhs
        @

    # Multpliy this `Matrix4` by a scalar value `rhs`, createing a new Matrix4.
    Matrix4::smul = (rhs) ->
        @copy().ismul rhs

    # Compute the determinant of this `Matrix4`.
    Matrix4::determinant = ->
        (@m[ 0] * @m[ 5] - @m[ 1] * @m[ 4]) *
        (@m[10] * @m[15] - @m[11] * @m[14]) -
        (@m[ 0] * @m[ 6] - @m[ 2] * @m[ 4]) *
        (@m[ 9] * @m[15] - @m[11] * @m[13]) +
        (@m[ 0] * @m[ 7] - @m[ 3] * @m[ 4]) *
        (@m[ 9] * @m[14] - @m[10] * @m[13]) +
        (@m[ 1] * @m[ 6] - @m[ 2] * @m[ 5]) *
        (@m[ 8] * @m[15] - @m[11] * @m[12]) -
        (@m[ 1] * @m[ 7] - @m[ 3] * @m[ 5]) *
        (@m[ 8] * @m[14] - @m[10] * @m[12]) +
        (@m[ 2] * @m[ 7] - @m[ 3] * @m[ 6]) *
        (@m[ 8] * @m[13] - @m[ 9] * @m[12])

    # Invert this `Matrix4` in place. It may throw a `ValueError` if the matrix
    # is degenerate.
    Matrix4::iinv = ->
        a0 = @m[ 0] * @m[ 5] - @m[ 1] * @m[ 4]
        a1 = @m[ 0] * @m[ 6] - @m[ 2] * @m[ 4]
        a2 = @m[ 0] * @m[ 7] - @m[ 3] * @m[ 4]
        a3 = @m[ 1] * @m[ 6] - @m[ 2] * @m[ 5]
        a4 = @m[ 1] * @m[ 7] - @m[ 3] * @m[ 5]
        a5 = @m[ 2] * @m[ 7] - @m[ 3] * @m[ 6]
        b0 = @m[ 8] * @m[13] - @m[ 9] * @m[12]
        b1 = @m[ 8] * @m[14] - @m[10] * @m[12]
        b2 = @m[ 8] * @m[15] - @m[11] * @m[12]
        b3 = @m[ 9] * @m[14] - @m[10] * @m[13]
        b4 = @m[ 9] * @m[15] - @m[11] * @m[13]
        b5 = @m[10] * @m[15] - @m[11] * @m[14]

        det = a0 * b5 - a1 * b4 + a2 * b3 + a3 * b2 - a4 * b1 + a5 * b0

        if (Math.abs det) < 1e-8
            throw new exceptions.RangeError "matrix is degenerate"

        [
            @m[ 0], @m[ 4], @m[ 8], @m[12],
            @m[ 1], @m[ 5], @m[ 9], @m[13],
            @m[ 2], @m[ 6], @m[10], @m[14],
            @m[ 3], @m[ 7], @m[11], @m[15]
        ] = [
            ( @m[ 5] * b5 - @m[ 6] * b4 + @m[ 7] * b3) / det,
            (-@m[ 4] * b5 + @m[ 6] * b2 - @m[ 7] * b1) / det,
            ( @m[ 4] * b4 - @m[ 5] * b2 + @m[ 7] * b0) / det,
            (-@m[ 4] * b3 + @m[ 5] * b1 - @m[ 6] * b0) / det,
            (-@m[ 1] * b5 + @m[ 2] * b4 - @m[ 3] * b3) / det,
            ( @m[ 0] * b5 - @m[ 2] * b2 + @m[ 3] * b1) / det,
            (-@m[ 0] * b4 + @m[ 1] * b2 - @m[ 3] * b0) / det,
            ( @m[ 0] * b3 - @m[ 1] * b1 + @m[ 2] * b0) / det,
            ( @m[13] * a5 - @m[14] * a4 + @m[15] * a3) / det,
            (-@m[12] * a5 + @m[14] * a2 - @m[15] * a1) / det,
            ( @m[12] * a4 - @m[13] * a2 + @m[15] * a0) / det,
            (-@m[12] * a3 + @m[13] * a1 - @m[14] * a0) / det,
            (-@m[ 9] * a5 + @m[10] * a4 - @m[11] * a3) / det,
            ( @m[ 8] * a5 - @m[10] * a2 + @m[11] * a1) / det,
            (-@m[ 8] * a4 + @m[ 9] * a2 - @m[11] * a0) / det,
            ( @m[ 8] * a3 - @m[ 9] * a1 + @m[10] * a0) / det
        ]
        @

    # Invert this `Matrix4` and output the inverse.
    Matrix4::inv = ->
        @copy().iinv()

    # Transpose this `Matrix4` (swap rows and columns) in-place.
    Matrix4::itranspose = ->
        [ @m[ 4], @m[ 1] ] = [ @m[ 1], @m[ 4] ]
        [ @m[ 8], @m[ 2] ] = [ @m[ 2], @m[ 8] ]
        [ @m[12], @m[ 3] ] = [ @m[ 3], @m[12] ]
        [ @m[ 9], @m[ 6] ] = [ @m[ 6], @m[ 9] ]
        [ @m[13], @m[ 7] ] = [ @m[ 7], @m[13] ]
        [ @m[14], @m[11] ] = [ @m[11], @m[14] ]
        @

    # Transpose this `Matrix4` and output the transposed.
    Matrix4::transpose = ->
        @copy().itranspose()

    # Decompose (if possible) this `Matrix4` into translation, scale and
    # rotation components. The translation and scale are `Vector3`s while the
    # rotation is a `Quaternion`.
    Matrix4::decompose = ->
        translation = new Vector3 @m[12], @m[13], @m[14]

        scale = new Vector3 \
            Math.sqrt(@m[ 0] * @m[ 0] + @m[ 4] * @m[ 4] + @m[ 8] * @m[ 8]),
            Math.sqrt(@m[ 1] * @m[ 1] + @m[ 5] * @m[ 5] + @m[ 9] * @m[ 9]),
            Math.sqrt(@m[ 2] * @m[ 2] + @m[ 6] * @m[ 6] + @m[10] * @m[10])

        translation : translation
        scale       : scale
        rotation    : Quaternion.fromRotationMatrix new Matrix4 new Matrix4.Array [
                @m[ 0] / scale.x, @m[ 1] / scale.y, @m[ 2] / scale.z, 0,
                @m[ 4] / scale.x, @m[ 5] / scale.y, @m[ 6] / scale.z, 0,
                @m[ 8] / scale.x, @m[ 9] / scale.y, @m[10] / scale.z, 0,
                               0,                0,                0, 1
            ]

    # Create a translation matrix from a `Vector3` `vec`.
    Matrix4.createTranslation = (vec) ->
        new Matrix4 new Matrix4.Array [
                1,      0,      0,      0,
                0,      1,      0,      0,
                0,      0,      1,      0,
            vec.x,  vec.y,  vec.z,      1
        ]

    # Create a scaling matrix from a `Vector3` `vec`.
    Matrix4.createScale = (vec) ->
        new Matrix4 new Matrix4.Array [
            vec.x,     0,     0,     0,
                0, vec.y,     0,     0,
                0,     0, vec.z,     0,
                0,     0,     0,     1
        ]

    # Create a rotation matrix from a `Quaternion` `q`.
    Matrix4.createRotation = (q) ->
        xx = q.x * q.x
        xy = q.x * q.y
        xz = q.x * q.z
        xw = q.x * q.w

        yy = q.y * q.y
        yz = q.y * q.z
        yw = q.y * q.w

        zz = q.z * q.z
        zw = q.z * q.w

        new Matrix4 new Matrix4.Array [
            1 - 2 * (yy + zz),     2 * (xy + zw),     2 * (xz - yw), 0,
                2 * (xy - zw), 1 - 2 * (xx + zz),     2 * (yz + xw), 0,
                2 * (xz + yw),     2 * (yz - xw), 1 - 2 * (xx + yy), 0,
                            0,                 0,                 0, 1
        ]

    # Compose a matrix from a `translation` `Vector3`, a `scale` `Vector3` and
    # a `rotation` `Quaternion`.
    Matrix4.compose = (translation, scale, rotation) ->
        Matrix4.createTranslation(translation)
            .immul(Matrix4.createScale(scale))
            .immul(Matrix4.createRotation(rotation))

    # Create a frustum matrix.
    Matrix4.createFrustum = (left, right, bottom, top, near, far) ->
        width = right - left
        height = top - bottom
        depth = far - near

        new Matrix4 new Matrix4.Array [
                  2 * near / width,                          0,                        0,  0,
                                 0,          2 * near / height,                        0,  0,
                                 0,                          0,    -(far + near) / depth, -1,
            (right + left) / width,    (top + bottom) / height,  -2 * far * near / depth,  0
        ]

    # Create a perspective matrix, specifying the field-of-view in degrees
    # `fovy`, the aspect ratio `aspect`, the near position `znear` and the far
    # position `zfar`.
    Matrix4.createPerspective = (fovy, aspect, near, far) ->
        tangent = Math.tan fovy * Math.PI / 360
        height = near * tangent
        width = height * aspect

        Matrix4.createFrustum -width, width, -height, height, near, far

    # Create an orthographic projective matrix consisting of the six clipping
    # planes `left`, `right`, `bottom`, `top`, `near` and `far`.
    Matrix4.createOrtho = (left, right, bottom, top, near, far) ->
        width = right - left
        height = top - bottom
        depth = far - near

        new Matrix4 new Matrix4.Array [
                         2 / width,                          0,                       0, 0,
                                 0,                 2 / height,                       0, 0,
                                 0,                          0,              -2 / depth, 0,
            -(right + left) / width,  -(top + bottom) / height,   -(far + near) / depth, 1
        ]

    # Create a "look at" matrix --- one which projects world space into camera
    # space --- using `eye` as the position of the viewer, `center` as the
    # position the viewer is looking at and `up` as the normalized "up" vector.
    Matrix4.createLookAt = (eye, center, up) ->
        fx = center.x - eye.x
        fy = center.y - eye.y
        fz = center.z - eye.z

        fnorm = Math.sqrt fx * fx + fy * fy + fz * fz

        if fnorm
            fx /= fnorm
            fy /= fnorm
            fz /= fnorm

        sx = fy * up.z - fz * up.y
        sy = fz * up.x - fx * up.z
        sz = fx * up.y - fy * up.x

        snorm = Math.sqrt sx * sx + sy * sy + sz * sz

        if snorm
            sx /= snorm
            sy /= snorm
            sz /= snorm

        ux = sy * fz - sz * fy
        uy = sz * fx - sx * fz
        uz = sx * fy - sy * fx

        new Matrix4 new Matrix4.Array [
             sx,  ux, -fx, 0,
             sy,  uy, -fy, 0,
             sz,  uz, -fz, 0,
            -(eye.x * sx + eye.y * sy + eye.z * sz), -(eye.x * ux + eye.y * uy + eye.z * uz), (eye.x * fx + eye.y * fy + eye.z * fz), 1
        ]

    # Create a matrix with ones along the leading diagonal, otherwise known as
    # the identity matrix.
    Matrix4.eye = ->
        new Matrix4 new Matrix4.Array [
            1, 0, 0, 0,
            0, 1, 0, 0,
            0, 0, 1, 0,
            0, 0, 0, 1
        ]

    module.exports = Matrix4
