# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This the Quaternion class, which represents a way to rotate objects in
# three-dimensional space.

define (require, exports, module) ->
    types = require './../../types'

    # A four-dimensional complex numbers as a system for rotations about a
    # given axis. The `x`, `y` and `z` parameters are the _i_, _j_ and _k_
    # coefficients respectively, and `w` is the real component. The
    # corresponding component can be accessed with one of `.x`, `.y`, `.z` or
    # `.w`.
    Quaternion = (x=0, y=0, z=0, w=0) ->
        @x = x
        @y = y
        @z = z
        @w = w
        return

    # Alias `types.Float32Array`.
    Quaternion.Array = types.Float32Array

    # Convert the `Quaternion` into a string representation.
    Quaternion::toString = ->
        "Quaternion(" + @x + ", " + @y + ", " + @z + ", " + @w + ")"

    # Copy the `Quaternion` into an array.
    Quaternion::toArray = ->
        new Quaternion.Array [ @x, @y, @z, @w ]

    # Copy the `Quaternion` into a new `Quaternion`.
    Quaternion::copy = ->
        new Quaternion @x, @y, @z, @w

    # Check if this `Quaternion` is exactly equal to the `rhs` `Quaternion`.
    Quaternion::equals = (rhs) ->
        !!(@x == rhs.x && @y == rhs.y && @z == rhs.z && @w == rhs.w)

    # Multiply this `Quaternion`, in-place, by the `rhs` `Quaternion`.
    Quaternion::imul = (rhs) ->
        [
            @x,
            @y,
            @z,
            @w
        ] = [
            @w * rhs.x + @x * rhs.w + @y * rhs.z - @z * rhs.y,
            @w * rhs.y + @y * rhs.w + @z * rhs.x - @x * rhs.z,
            @w * rhs.z + @z * rhs.w + @x * rhs.y - @y * rhs.x,
            @w * rhs.w - @x * rhs.x - @y * rhs.y - @z * rhs.z
        ]
        @

    # Multiply this `Quaternion` with another `Quaternion`.
    Quaternion::mul = (rhs) ->
        @copy().imul rhs

    # Calculate the Euclidean norm squared of this `Quaternion`.
    Quaternion::normSquared = ->
        @x * @x + @y * @y + @z * @z + @w * @w

    # Calculate the Euclidean norm of this `Quaternion`.
    Quaternion::norm = ->
        Math.sqrt @x * @x + @y * @y + @z * @z + @w * @w

    # Divide each component of the `Quaternion` by the norm, so that the new
    # norm the quaternion is 1. This operation is in-place.
    Quaternion::normalize = ->
        norm = Math.sqrt @x * @x + @y * @y + @z * @z + @w * @w

        @x /= norm
        @y /= norm
        @z /= norm
        @w /= norm
        @

    # Invert the `Quaternion`, in-place.
    Quaternion::iinv = ->
        normSquared = @x * @x + @y * @y + @z * @z + @w * @w

        @x = -@x / normSquared
        @y = -@y / normSquared
        @z = -@z / normSquared
        @w =  @w / normSquared
        @

    # Invert the `Quaternion`.
    Quaternion::inv = ->
        @copy().iinv()

    # Perform a spherical-linear interpolation of the `Quaternion` with another
    # `Quaternion` `rhs`, in-place, by the amount `amt`.
    Quaternion::islerp = (rhs, amt) ->
        @x += (rhs.x - @x) * amt
        @y += (rhs.y - @y) * amt
        @z += (rhs.z - @z) * amt
        @w += (rhs.w - @w) * amt
        @

    # Perform a spherical-linear interpolation of the `Quaternion` with another
    # `Quaternion` `rhs`, by the amount `amt`.
    Quaternion::slerp = (rhs, amt) ->
        @copy().islerp rhs, amt

    # Create a `Quaternion` from an axis-angle pair --- the `axis` being a
    # `Vector3` and the `angle` being a scalar.
    Quaternion.fromAxisAngle = (axis, angle) ->
        (new Quaternion \
            axis.x * Math.sin(angle / 2),
            axis.y * Math.sin(angle / 2),
            axis.z * Math.sin(angle / 2),
            Math.cos(angle / 2)
        ).normalize()

    # Attempt to extract a `Quaternion` from a rotation `Matrix4` `mat`. It
    # will also work on matrices that are not rotational matrices, but the
    # results are undefined.
    Quaternion.fromRotationMatrix = (mat) ->
        trace = mat.m[ 0] + mat.m[ 5] + mat.m[10] + 1

        if trace > 0
            s = 0.5 / Math.sqrt trace
            new Quaternion \
                (mat.m[ 6] - mat.m[ 9]) * s,
                (mat.m[ 8] - mat.m[ 2]) * s,
                (mat.m[ 1] - mat.m[ 4]) * s,
                0.25 / s

        else if mat.m[ 0] > mat.m[ 5] && mat.m[ 0] > mat.m[10]
            s = Math.sqrt(1 + mat.m[ 0] - mat.m[ 5] - mat.m[10]) * 2
            new Quaternion \
                0.5 / s,
                (mat.m[ 4] + mat.m[ 1]) / s,
                (mat.m[ 8] + mat.m[ 2]) / s,
                (mat.m[ 9] + mat.m[ 6]) / s

        else if mat.m[ 5] > mat.m[ 0] && mat.m[ 5] > mat.m[10]
            s = Math.sqrt(1 + mat.m[ 5] - mat.m[ 0] - mat.m[10]) * 2
            new Quaternion \
                (mat.m[ 4] + mat.m[ 1]) / s,
                0.5 / s,
                (mat.m[ 9] + mat.m[ 6]) / s,
                (mat.m[ 8] + mat.m[ 2]) / s

        else if mat.m[10] > mat.m[ 0] && mat.m[10] > mat.m[ 5]
            s = Math.sqrt(1 + mat.m[10] - mat.m[ 0] - mat.m[ 5]) * 2
            new Quaternion \
                (mat.m[ 8] + mat.m[ 2]) / s,
                (mat.m[ 9] + mat.m[ 6]) / s,
                0.5 / s,
                (mat.m[ 4] + mat.m[ 1]) / s

    module.exports = Quaternion
