# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is the Vector3 class, which represents a 3-dimensional vector.

define (require, exports, module) ->
    types = require './../../types'

    # A representation of a three-dimensional vector. Takes the parameters `x`,
    # `y` and `z` representing the x, y and z components, respectively.
    #
    # The parameters are accessible as `.x`, `.y` and `.z`.
    Vector3 = (x=0, y=0, z=0) ->
        @x = x
        @y = y
        @z = z
        return

    # Alias `types.Float32Array`.
    Vector3.Array = types.Float32Array

    # Convert the `Vector3` into a string representation.
    Vector3::toString = ->
        "Vector3(" + @x + ", " + @y + ", " + @z + ")"

    # Copy the `Vector3` into an array.
    Vector3::toArray = ->
        new Vector3.Array [ @x, @y, @z ]

    # Copy the `Vector3` into another `Vector3`. The `copy` method can be used
    # to prevent mutation of vectors by creating a copy before applying any
    # in-place operations on them.
    Vector3::copy = ->
        new Vector3 @x, @y, @z

    # Check if two `Vector3`s are exactly equal.
    Vector3::equals = (rhs) ->
        !!(@x == rhs.x and @y == rhs.y and @z = rhs.z)

    # Add to another `Vector3` in-place, using the parameter `rhs` as the other
    # vector.
    Vector3::iadd = (rhs) ->
        @x += rhs.x
        @y += rhs.y
        @z += rhs.z
        @

    # Add this `Vector3` to another `Vector3`.
    Vector3::add = (rhs) ->
        @copy().iadd rhs

    # Subtract this `Vector3` from another `Vector3` in-place, using the
    # parameter `rhs` as the other vector.
    Vector3::isub = (rhs) ->
        @x -= rhs.x
        @y -= rhs.y
        @z -= rhs.z
        @

    # Subtract this `Vector3` from another `Vector3`.
    Vector3::sub = (rhs) ->
        @copy().isub rhs

    # Multiply this `Vector3` with something in-place.
    Vector3::imul = (rhs) ->
        if typeof rhs == 'number'
            @ismul rhs
        else
            @ivmul rhs

    # Multiply this `Vector3` with something.
    Vector3::mul = (rhs) ->
        @copy().imul rhs

    # Multiply this `Vector3` with another `Vector3` in-place, using the
    # parameter `rhs` as the other vector.
    Vector3::ivmul = (rhs) ->
        @x *= rhs.x
        @y *= rhs.y
        @z *= rhs.z
        @

    # Multiply this `Vector3` with another `Vector3`.
    Vector3::vmul = (rhs) ->
        @copy().ivmul rhs

    # Multiply this `Vector3` with another scalar in-place, using the parameter
    # `rhs` as the scalar.
    Vector3::ismul = (rhs) ->
        @x *= rhs
        @y *= rhs
        @z *= rhs
        @

    # Multiply this `Vector3` with another scalar.
    Vector3::smul = (rhs) ->
        @copy().ismul rhs

    # Compute the dot product between this `Vector3` and another `Vector3`
    # `rhs`.
    Vector3::dot = (rhs) ->
        @x * rhs.x + @y * rhs.y + @z * rhs.z

    # Cross this `Vector3` with another `Vector3` `rhs` in-place.
    Vector3::icross = (rhs) ->
        [
            @x,
            @y,
            @z
        ] = [
            @y * rhs.z - @z * rhs.y,
            @z * rhs.x - @x * rhs.z,
            @x * rhs.y - @y * rhs.x
        ]
        @

    # Cross this `Vector3` with another `Vector3`.
    Vector3::cross = (rhs) ->
        @copy().icross rhs

    # Calculate the Euclidean norm (magnitude) squared. This is cheaper than
    # calculating the actual norm if only relative comparisons are needed.
    Vector3::normSquared = ->
        @x * @x + @y * @y + @z * @z

    # Calculate the Euclidean norm (magnitude). This is expensive and if only
    # relative distance comparisons are required, it is more efficient to use
    # `Vector3::normSquared`.
    Vector3::norm = ->
        Math.sqrt @x * @x + @y * @y + @z * @z

    # Normalize the `Vector3` in-place (dividing it by the norm so the new norm
    # is 1).
    Vector3::normalize = ->
        norm = Math.sqrt(@x * @x + @y * @y + @z * @z)

        @x /= norm
        @y /= norm
        @z /= norm
        @

    # Transform the `Vector3` in-place using a `Matrix4` `mat`.
    Vector3::itransform = (mat) ->
        w = mat.m[ 3] * @x + mat.m[ 7] * @y + mat.m[11] * @z + mat.m[15]

        [
            @x,
            @y,
            @z
        ] = [
            (mat.m[ 0] * @x + mat.m[ 4] * @y + mat.m[ 8] * @z + mat.m[12]) / w,
            (mat.m[ 1] * @x + mat.m[ 5] * @y + mat.m[ 9] * @z + mat.m[13]) / w,
            (mat.m[ 2] * @x + mat.m[ 6] * @y + mat.m[10] * @z + mat.m[14]) / w
        ]
        @

    # Transform the `Vector3` using a `Matrix4` `mat`.
    Vector3::transform = (mat) ->
        @copy().itransform mat

    # Lerp this `Vector3` in-place with another `Vector3` `rhs` by the amount
    # `amt`.
    Vector3::ilerp = (rhs, amt) ->
        @x += (rhs.x - @x) * amt
        @y += (rhs.y - @y) * amt
        @z += (rhs.z - @z) * amt
        @

    # Lerp this `Vector3` with another `Vector3` `rhs` by the amount `amt`.
    Vector3::lerp = (rhs, amt) ->
        @copy().ilerp rhs, amt

    module.exports = Vector3
