# Copyright 2011 Tony Young
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

define (require, exports, module) ->
    Matrix4 = require 'basilisk/math/linalg/Matrix4'
    Quaternion = require 'basilisk/math/linalg/Quaternion'
    Vector3 = require 'basilisk/math/linalg/Vector3'

    window.module 'basilisk/math/linalg/Matrix4'

    TOLERANCE = 0.00001

    _fpMatrixEqual = (act, exp, tolerance) ->
        for i in [0..15]
            ok Math.abs(act.m[i] - exp.m[i]) < tolerance, 'm[' + i + '] (floating point comparison)'

    test 'Matrix4.fromRowMajor', ->
        mat = Matrix4.fromRowMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16

        deepEqual mat, new Matrix4 new Matrix4.Array [
             1,  5,  9, 13,
             2,  6, 10, 14,
             3,  7, 11, 15,
             4,  8, 12, 16
        ]

    test 'Matrix4.fromColumnMajor', ->
        mat = Matrix4.fromColumnMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16

        deepEqual mat, new Matrix4 new Matrix4.Array [
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16
        ]

    test 'Matrix4::toString', ->
        mat = Matrix4.fromRowMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16
        
        equals mat.toString(), """
         Matrix4([1, 5, 9, 13,
\ \ \ \ \ \ \ \ \ 2, 6, 10, 14,
\ \ \ \ \ \ \ \ \ 3, 7, 11, 15,
\ \ \ \ \ \ \ \ \ 4, 8, 12, 16])""", 'toString'

    test 'Matrix4::equals', ->
        mat1 = Matrix4.fromRowMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16
        
        mat2 = Matrix4.fromRowMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16
        
        equal mat1.equals(mat2), true, 'equal'

        mat2 = Matrix4.fromRowMajor \
             4,  2, 32,  3,
             2,  1,  2,  2,
             3,  4,  0,  6,
             6,  3,  5,  7
        
        equal mat1.equals(mat2), false, 'not equal'

    test 'Matrix4::add', ->
        mat1 = Matrix4.fromRowMajor \
            1, 2, 3, 4,
            3, 3, 2, 1,
            2, 4, 1, 2,
            4, 1, 2, 3

        mat2 = Matrix4.fromRowMajor \
            1, 4, 3, 2,
            3, 2, 1, 3,
            2, 1, 4, 3,
            4, 3, 2, 1

        exp = Matrix4.fromRowMajor \
            2, 6, 6, 6,
            6, 5, 3, 4,
            4, 5, 5, 5,
            8, 4, 4, 4

        out = mat1.add mat2

        deepEqual out, exp

    test 'Matrix4::mul', ->
        mat1 = Matrix4.fromRowMajor \
            1, 2, 3, 4,
            3, 3, 2, 1,
            2, 4, 1, 2,
            4, 1, 2, 3

        mat2 = Matrix4.fromRowMajor \
            1, 4, 3, 2,
            3, 2, 1, 3,
            2, 1, 4, 3,
            4, 3, 2, 1

        exp = Matrix4.fromRowMajor \
            29, 23, 25, 21,
            20, 23, 22, 22,
            24, 23, 18, 21,
            23, 29, 27, 20

        out = mat1.mul mat2

        deepEqual out, exp, 'mmul'

        mat = Matrix4.fromRowMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16

        out = mat.mul 5

        exp = Matrix4.fromRowMajor \
             5, 10, 15, 20,
            25, 30, 35, 40,
            45, 50, 55, 60,
            65, 70, 75, 80

        deepEqual out, exp, 'smul'

    test 'Matrix4::mmul', ->
        mat1 = Matrix4.fromRowMajor \
            1, 2, 3, 4,
            3, 3, 2, 1,
            2, 4, 1, 2,
            4, 1, 2, 3

        mat2 = Matrix4.fromRowMajor \
            1, 4, 3, 2,
            3, 2, 1, 3,
            2, 1, 4, 3,
            4, 3, 2, 1

        exp = Matrix4.fromRowMajor \
            29, 23, 25, 21,
            20, 23, 22, 22,
            24, 23, 18, 21,
            23, 29, 27, 20

        out = mat1.mmul mat2

        deepEqual out, exp

    test 'Matrix4::smul', ->
        mat = Matrix4.fromRowMajor \
             1,  2,  3,  4,
             5,  6,  7,  8,
             9, 10, 11, 12,
            13, 14, 15, 16

        out = mat.smul 5

        exp = Matrix4.fromRowMajor \
             5, 10, 15, 20,
            25, 30, 35, 40,
            45, 50, 55, 60,
            65, 70, 75, 80

        deepEqual out, exp

    test 'Matrix4::determinant', ->
        mat = Matrix4.fromRowMajor \
            1, 4, 3, 2,
            3, 2, 1, 3,
            2, 1, 4, 3,
            4, 3, 2, 1
        
        ok Math.abs(mat.determinant() - 116) < TOLERANCE, 'determinant (floating point comparison)'

    test 'Matrix4::inv', ->
        mat = Matrix4.fromRowMajor \
            1, 4, 3, 2,
            3, 2, 1, 3,
            2, 1, 4, 3,
            4, 3, 2, 1
        
        exp = Matrix4.fromRowMajor \
            -0.22414,  0.03448,  0.02586,  0.26724,
             0.27586,  0.03448, -0.22414,  0.01724,
             0.01724, -0.31034,  0.26724,  0.09483,
             0.03448,  0.37931,  0.03448, -0.31034

        _fpMatrixEqual mat.inv(), exp, TOLERANCE

        mat = Matrix4.fromRowMajor \
            1, 2, 3, 4,
            5, 6, 7, 8,
            9, 10, 11, 12,
            13, 14, 15, 16
        
        try
            mat.inv()
            ok false, 'degenerate matrix inverted'
        catch e
            ok true, 'degenerate'

    test 'Matrix4::transpose', ->
        mat = Matrix4.fromRowMajor \
            1, 4, 3, 2,
            3, 2, 1, 3,
            2, 1, 4, 3,
            4, 3, 2, 1
        
        matt = Matrix4.fromRowMajor \
            1, 3, 2, 4,
            4, 2, 1, 3,
            3, 1, 4, 2,
            2, 3, 3, 1
        
        ok mat.transpose().equals(matt), 'transpose'

    test 'Matrix4::decompose/Matrix4.compose', ->
        scale = new Vector3 1, 2, 3
        rotation = (new Quaternion 1, 2, 3, 4).normalize()
        translation = new Vector3 1, 2, 3

        mat = Matrix4.compose translation, scale, rotation
        decomp = mat.decompose()

        ok Math.abs(scale.x - decomp.scale.x) < TOLERANCE, 'x component'
        ok Math.abs(scale.y - decomp.scale.y) < TOLERANCE, 'y component'
        ok Math.abs(scale.z - decomp.scale.z) < TOLERANCE, 'z component'
        ok Math.abs(rotation.x - decomp.rotation.x) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(rotation.y - decomp.rotation.y) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(rotation.z - decomp.rotation.z) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(rotation.w - decomp.rotation.w) < TOLERANCE, 'real component (floating point comparison)'
        ok Math.abs(translation.x - decomp.translation.x) < TOLERANCE, 'x component'
        ok Math.abs(translation.y - decomp.translation.y) < TOLERANCE, 'y component'
        ok Math.abs(translation.z - decomp.translation.z) < TOLERANCE, 'z component'

    test 'Matrix4.createTranslation', ->
        mat = Matrix4.createTranslation \
            new Vector3(1, 2, 3)
        
        ok mat.equals Matrix4.fromRowMajor \ 
            1, 0, 0, 1,
            0, 1, 0, 2,
            0, 0, 1, 3,
            0, 0, 0, 1

    test 'Matrix4.createScale', ->
        mat = Matrix4.createScale \ 
            new Vector3(1, 2, 3)
        
        ok mat.equals Matrix4.fromRowMajor \
            1, 0, 0, 0,
            0, 2, 0, 0,
            0, 0, 3, 0,
            0, 0, 0, 1

    test 'Matrix4.createRotation', ->
        mat = Matrix4.createRotation \
            new Quaternion(1, 2, 3, 4).normalize()
        
        exp = Matrix4.fromRowMajor \
             0.13333, -0.66667, 0.73333, 0,
             0.93333,  0.33333, 0.13333, 0,
            -0.33333,  0.66667, 0.66667, 0,
                   0,        0,       0, 1
        
        _fpMatrixEqual mat, exp, TOLERANCE

    test 'Matrix4.createFrustum', ->
        mat = Matrix4.createFrustum \
            -100,
            100,
            -100,
            100,
            -100,
            100
        
        exp = Matrix4.fromRowMajor \
            -1,  0,  0,    0,
             0, -1,  0,    0,
             0,  0,  0,  100,
             0,  0, -1,    0

        _fpMatrixEqual mat, exp, TOLERANCE

    test 'Matrix4.createPerspective', ->
        mat = Matrix4.createPerspective \
            45,
            0.5,
            1,
            100
        
        exp = Matrix4.fromRowMajor \
             4.82843,        0,        0,        0,
                   0,  2.41421,        0,        0,
                   0,        0, -1.02020, -2.02020,
                   0,        0,       -1,        0
        
        _fpMatrixEqual mat, exp, TOLERANCE

    test 'Matrix4.createOrtho', ->
        mat = Matrix4.createOrtho \
            -100,
            100,
            -100,
            100,
            -100,
            100

        exp = Matrix4.fromRowMajor \
                0.01,        0,        0,        0,
                   0,     0.01,        0,        0,
                   0,        0,    -0.01,        0,
                   0,        0,        0,        1

        _fpMatrixEqual mat, exp, TOLERANCE

    test 'Matrix4.createLookAt', ->
        mat = Matrix4.createLookAt \
            new Vector3(0, 0, 0),
            new Vector3(1, 0, 0),
            new Vector3(0, 1, 0)
        
        exp = Matrix4.fromRowMajor \
             0,  0,  1,  0,
             0,  1,  0,  0,
            -1,  0,  0,  0,
             0,  0,  0,  1
        
        _fpMatrixEqual mat, exp, TOLERANCE

    test 'Matrix4.eye', ->
        mat = Matrix4.eye()
        
        exp = Matrix4.fromRowMajor \
             1,  0,  0,  0,
             0,  1,  0,  0,
             0,  0,  1,  0,
             0,  0,  0,  1
        
        deepEqual mat, exp
