# Copyright 2011 Tony Young
#                David Czech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

define (require, exports, module) ->
    Quaternion = require 'basilisk/math/linalg/Quaternion'
    Matrix4 = require 'basilisk/math/linalg/Matrix4'
    Vector3 = require 'basilisk/math/linalg/Vector3'

    window.module 'basilisk/math/linalg/Quaternion'

    TOLERANCE = 0.00001

    _fpMatrixEqual = (act, exp, tolerance) ->
        for i in [0..15]
            ok Math.abs(act.m[i] - exp.m[i]) < tolerance, 'm[' + i + '] (floating point comparison)'

    test 'new Quaternion', ->
        q = new Quaternion 1, 2, 3, 4
        equals q.x, 1, 'i coefficient'
        equals q.y, 2, 'j coefficient'
        equals q.z, 3, 'k coefficient'
        equals q.w, 4, 'real component'

    test 'Quaternion::toString', ->
        q = new Quaternion 1, 2, 3, 4
        equals q.toString(), 'Quaternion(1, 2, 3, 4)', 'toString'

    test 'Quaternion::toArray', ->
        q = new Quaternion 1, 2, 3, 4
        qa = q.toArray()
        equals qa[0], 1, 'i coefficient'
        equals qa[1], 2, 'j coefficient'
        equals qa[2], 3, 'k coefficient'
        equals qa[3], 4, 'real component'

    test 'Quaternion::copy', ->
        q1 = new Quaternion 1, 2, 3, 4
        ok (q1.copy().equals q1), 'equal'

    test 'Quaternion::equals', ->
        q1 = new Quaternion 1, 2, 3, 4
        q2 = new Quaternion 1, 2, 3, 4
        ok q1.equals q2, 'equal'
        q2 = new Quaternion 4, 5, 6, 7
        ok (!q1.equals q2), 'not equal'

    test 'Quaternion::mul', ->
        q1 = new Quaternion 1, 2, 3, 4
        q2 = new Quaternion 5, 6, 7, 8
        qo = q1.mul q2
        equals qo.x, 24, 'i coefficient'
        equals qo.y, 48, 'j coefficient'
        equals qo.z, 48, 'k coefficient'
        equals qo.w, -6, 'real component'

    test 'Quaternion::normSquared', ->
        q = new Quaternion 1, 2, 3, 4
        equals q.normSquared(), 30, 'normSquared'

    test 'Quaternion::norm', ->
        q = new Quaternion 1, 2, 3, 4
        equals q.norm(), Math.sqrt 30, 'norm'

    test 'Quaternion::normalize', ->
        q = new Quaternion 1, 2, 3, 4
        qn = q.copy().normalize()
        norm = q.norm()
        ok Math.abs(qn.norm() - 1) < TOLERANCE, 'norm (floating point comparison)'
        ok Math.abs(q.x / norm - qn.x) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(q.y / norm - qn.y) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(q.z / norm - qn.z) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(q.w / norm - qn.w) < TOLERANCE, 'real component (floating point comparison)'

    test 'Quaternion::inv', ->
        q = new Quaternion 1, 2, 3, 4
        qi = q.copy().inv()
        normSquared = q.normSquared()
        ok Math.abs(qi.x + 1 / normSquared) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(qi.y + 2 / normSquared) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(qi.z + 3 / normSquared) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(qi.w - 4 / normSquared) < TOLERANCE, 'real component (floating point comparison)'

    test 'Quaternion.fromRotationMatrix', ->
        q1 = (new Quaternion 1, 2, 3, 4).normalize()
        qo = Quaternion.fromRotationMatrix Matrix4.createRotation q1
        ok Math.abs(qo.x - q1.x) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(qo.y - q1.y) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(qo.z - q1.z) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(qo.w - q1.w) < TOLERANCE, 'real component (floating point comparison)'

        q1 = (new Quaternion 4, 3, 2, 1).normalize()
        qo = Quaternion.fromRotationMatrix Matrix4.createRotation q1
        ok Math.abs(qo.x - q1.x) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(qo.y - q1.y) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(qo.z - q1.z) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(qo.w - q1.w) < TOLERANCE, 'real component (floating point comparison)'

        mat = Matrix4.fromRowMajor \
            1, 0, 0, 0,
            0,-1, 0, 0,
            0, 0,-1, 0,
            0, 0, 0, 1
        
        qo = Quaternion.fromRotationMatrix(mat).normalize()
        exp = Matrix4.createRotation qo

        _fpMatrixEqual mat, exp, TOLERANCE

        mat = Matrix4.fromRowMajor \
           -1, 0, 0, 0,
            0, 1, 0, 0,
            0, 0,-1, 0,
            0, 0, 0, 1
        
        qo = Quaternion.fromRotationMatrix(mat).normalize()
        exp = Matrix4.createRotation qo

        _fpMatrixEqual mat, exp, TOLERANCE

        mat = Matrix4.fromRowMajor \
           -1, 0, 0, 0,
            0,-1, 0, 0,
            0, 0, 1, 0,
            0, 0, 0, 1
        
        qo = Quaternion.fromRotationMatrix(mat).normalize()
        exp = Matrix4.createRotation qo

        _fpMatrixEqual mat, exp, TOLERANCE

    test 'Quaternion.fromAxisAngle', ->
        q = new Quaternion.fromAxisAngle (new Vector3 1, 2, 3), 10
        ok Math.abs(q.x + 0.266430) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(q.y + 0.532860) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(q.z + 0.799290) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(q.w - 0.078813) < TOLERANCE, 'real component (floating point comparison)'

    test 'Quaternion::slerp', ->
        q1 = new Quaternion 1, 2, 3, 4
        q2 = new Quaternion 5, 6, 7, 8
        qo = q1.slerp q2, 0.5
        ok Math.abs(qo.x - 3) < TOLERANCE, 'i coefficient (floating point comparison)'
        ok Math.abs(qo.y - 4) < TOLERANCE, 'j coefficient (floating point comparison)'
        ok Math.abs(qo.z - 5) < TOLERANCE, 'k coefficient (floating point comparison)'
        ok Math.abs(qo.w - 6) < TOLERANCE, 'real component (floating point comparison)'
